==> xcerdisp-1.13pre.tar.gz: xcerdisp-1.13pre/lru.c <==
/* $Id: lru.c,v 1.7 2002/08/09 01:16:18 d Exp $ */
/* David Leonard, 2001. Public domain. */

/*
 * LRU colour management. Because the author has a 8bpp display
 * he's very interested in robust and best-effort colour allocation.
 * This is simply overkill on displays with a depth of 15 or more,
 * but its harmless.
 */

#include <X11/IntrinsicP.h>
#include <X11/Xmd.h>

#include "lru.h"

struct lru {
    struct lru *next;
    struct lru **prevp;
    CARD16 col;
    unsigned long pixel;
};

struct lrulist {
    struct lru *mem;         /* LRU list storage */
    struct lru *free;        /* free list */
    struct lru *head;        /* allocated colours */
    struct lru **old;        /* youngest unused colour pointer */
    Widget widget;
};

static void lruinsert(struct lru **, struct lru *);
static void lruremove(struct lru *);

/* Insert a LRU element into the head of a list */
static void
lruinsert(struct lru **head, struct lru *lru)
{
    lru->next = *head;
    if (lru->next)
	lru->next->prevp = &lru->next;
    lru->prevp = head;
    *head = lru;
}

/* Remove a LRU element from a list */
static void
lruremove(struct lru *lru)
{
    *(lru->prevp) = lru->next;
    if (lru->next)
	lru->next->prevp = lru->prevp;
}

/* Allocate storage and a freelist of LRU structures */
struct lrulist *
lrualloc(Widget w)
{
    CARD16 i;
    struct lrulist *list;

    list = (struct lrulist *)XtMalloc(sizeof (struct lrulist));

    list->widget = w;
    list->mem = (struct lru *)XtMalloc(32*32*32*sizeof (struct lru));
    list->free = NULL;
    list->head = NULL;
    for (i = 0; i < 32*32*32; i++) 
	lruinsert(&list->free, &list->mem[i]);
    list->old = &list->head;
    return list;
}

/* Free all storage and colours associated with an LRU list */
void
lrufree(struct lrulist *list)
{
    struct lru *lru;
    for (lru = list->head; lru; lru = lru->next)
	XFreeColors(XtDisplay(list->widget), list->widget->core.colormap,
	    &lru->pixel, 1, 0);
    XtFree((char *)list->mem);
    XtFree((char *)list);
}

/* Free colours used older than old,
 * and reset old to the head of used colours */
void
lrufreeold(struct lrulist *list)
{
    while (*list->old) {
        struct lru *lru = *list->old;
	lruremove(lru);
	XFreeColors(XtDisplay(list->widget), list->widget->core.colormap,
	    &lru->pixel, 1, 0);
	lruinsert(&list->free, lru);
    }
    list->old = &list->head;
}

/* Find and allocate a pixel for the corresponding RGB colour 
 * using LRU managment */
unsigned long
lrupixel(struct lrulist *list, u_int16_t r, u_int16_t g, u_int16_t b)
{
    XColor xcolor;
    CARD16 col;
    struct lru *lru, *lrulast = NULL;
    int status;

    col = ((r & 0x1f) << 10) | ((g & 0x1f) << 5) | (b & 0x1f);

    if (col != 0) col |= 1;	/* XXX hack to reduce nr of blues */

    /* Has this colour already been allocated? */
    for (lru = list->head; lru; lrulast = lru, lru = lru->next)
	if (lru->col == col)
		break;
    /* If already allocated, move to head of the list */
    if (lru) {
	if (list->old == &lru->next)
	    list->old = lru->prevp;
	lruremove(lru);
	lruinsert(&list->head, lru);
	return lru->pixel;
    }

    /* Time to allocate a new colour */
    xcolor.red = (r == 0x1f) ? 65535 : (unsigned long)r << (16-5);
    xcolor.green = (g == 0x1f) ? 65535 : (unsigned long)g << (16-5);
    xcolor.blue = (b == 0x1f) ? 65535 : (unsigned long)b << (16-5);
    xcolor.flags = DoRed|DoGreen|DoBlue;

    /* 
     * XXX - XAllocColor requires a query to the server and recv a reply.
     * Expose and input events generated while we are allocating colours 
     * are not handled as rapidly as we would like, so this is quite
     * an expensive operation! Well, it seems to be on my server anyway.
     */
    status = XAllocColor(XtDisplay(list->widget), list->widget->core.colormap, &xcolor);

    if (!status) {
	struct lru *bestlru = NULL;
	long bestdist = 32*32*3;

	/* Could not alloc colour; search for closest existing colour */
	for (lru = list->head; lru; lru = lru->next) {
	    CARD16 or = (lru->col >> 10);
	    CARD16 og = (lru->col >> 5) & 0x1f;
	    CARD16 ob = (lru->col) & 0x1f;
	    long dist = (r-or)*(r-or)+(g-og)*(g-og)+(b-ob)*(b-ob);
	    if (dist < bestdist) {
		bestdist = dist;
		bestlru = lru;
	    }
	}
	if (!bestlru) 
	    return 0;	/* eek! */
	lru = bestlru;
	if (list->old == &lru->next)
	    list->old = lru->prevp;
	lruremove(lru);
	lruinsert(&list->head, lru);
	return lru->pixel;
    }

    /* Allocated new colour */
    lru = list->free;
    lruremove(lru);
    lru->col = col;
    lru->pixel = xcolor.pixel;
    lruinsert(&list->head, lru);
    return lru->pixel;
}

==> xcerdisp-1.14.1089.tar.gz: xcerdisp-1.14.1089/lru.c <==
/* $Id: lru.c 504 2003-08-26 05:50:48Z d $ */
/* David Leonard, 2001. Public domain. */

/*
 * LRU colour management. Because David has a 8bpp display
 * he's very interested in robust and best-effort colour allocation.
 * This is simply overkill on displays with a depth of 15 or more,
 * but its harmless.
 */

#include <X11/IntrinsicP.h>
#include <X11/Xmd.h>

#include "lru.h"

struct lru {
    struct lru *next;
    struct lru **prevp;
    CARD16 col;
    unsigned long pixel;
};

struct lrulist {
    struct lru *mem;         /* LRU list storage */
    struct lru *free;        /* free list */
    struct lru *head;        /* allocated colours */
    struct lru **old;        /* youngest unused colour pointer */
    Widget widget;
};

static void lruinsert(struct lru **, struct lru *);
static void lruremove(struct lru *);

/* Insert a LRU element into the head of a list */
static void
lruinsert(struct lru **head, struct lru *lru)
{
    lru->next = *head;
    if (lru->next)
	lru->next->prevp = &lru->next;
    lru->prevp = head;
    *head = lru;
}

/* Remove a LRU element from a list */
static void
lruremove(struct lru *lru)
{
    *(lru->prevp) = lru->next;
    if (lru->next)
	lru->next->prevp = lru->prevp;
}

/* Allocate storage and a freelist of LRU structures */
struct lrulist *
lrualloc(Widget w)
{
    CARD16 i;
    struct lrulist *list;

    list = (struct lrulist *)XtMalloc(sizeof (struct lrulist));

    list->widget = w;
    list->mem = (struct lru *)XtMalloc(32*32*32*sizeof (struct lru));
    list->free = NULL;
    list->head = NULL;
    for (i = 0; i < 32*32*32; i++) 
	lruinsert(&list->free, &list->mem[i]);
    list->old = &list->head;
    return list;
}

/* Free all storage and colours associated with an LRU list */
void
lrufree(struct lrulist *list)
{
    struct lru *lru;
    for (lru = list->head; lru; lru = lru->next)
	XFreeColors(XtDisplay(list->widget), list->widget->core.colormap,
	    &lru->pixel, 1, 0);
    XtFree((char *)list->mem);
    XtFree((char *)list);
}

/* Free colours used older than old,
 * and reset old to the head of used colours */
void
lrufreeold(struct lrulist *list)
{
    while (*list->old) {
        struct lru *lru = *list->old;
	lruremove(lru);
	XFreeColors(XtDisplay(list->widget), list->widget->core.colormap,
	    &lru->pixel, 1, 0);
	lruinsert(&list->free, lru);
    }
    list->old = &list->head;
}

/* Find and allocate a pixel for the corresponding RGB colour 
 * using LRU managment */
unsigned long
lrupixel(struct lrulist *list, CARD16 r, CARD16 g, CARD16 b)
{
    XColor xcolor;
    CARD16 col;
    struct lru *lru, *lrulast = NULL;
    int status;

    col = ((r & 0x1f) << 10) | ((g & 0x1f) << 5) | (b & 0x1f);

    if (col != 0) col |= 1;	/* XXX hack to reduce nr of blues */

    /* Has this colour already been allocated? */
    for (lru = list->head; lru; lrulast = lru, lru = lru->next)
	if (lru->col == col)
		break;
    /* If already allocated, move to head of the list */
    if (lru) {
	if (list->old == &lru->next)
	    list->old = lru->prevp;
	lruremove(lru);
	lruinsert(&list->head, lru);
	return lru->pixel;
    }

    /* Time to allocate a new colour */
    xcolor.red = (r == 0x1f) ? 65535 : (unsigned long)r << (16-5);
    xcolor.green = (g == 0x1f) ? 65535 : (unsigned long)g << (16-5);
    xcolor.blue = (b == 0x1f) ? 65535 : (unsigned long)b << (16-5);
    xcolor.flags = DoRed|DoGreen|DoBlue;

    /* 
     * XXX - XAllocColor requires a query to the server and recv a reply.
     * Expose and input events generated while we are allocating colours 
     * are not handled as rapidly as we would like, so this is quite
     * an expensive operation! Well, it seems to be on my server anyway.
     */
    status = XAllocColor(XtDisplay(list->widget), list->widget->core.colormap, &xcolor);

    if (!status) {
	struct lru *bestlru = NULL;
	long bestdist = 32*32*3;

	/* Could not alloc colour; search for closest existing colour */
	for (lru = list->head; lru; lru = lru->next) {
	    CARD16 or = (lru->col >> 10);
	    CARD16 og = (lru->col >> 5) & 0x1f;
	    CARD16 ob = (lru->col) & 0x1f;
	    long dist = (r-or)*(r-or)+(g-og)*(g-og)+(b-ob)*(b-ob);
	    if (dist < bestdist) {
		bestdist = dist;
		bestlru = lru;
	    }
	}
	if (!bestlru) 
	    return 0;	/* eek! */
	lru = bestlru;
	if (list->old == &lru->next)
	    list->old = lru->prevp;
	lruremove(lru);
	lruinsert(&list->head, lru);
	return lru->pixel;
    }

    /* Allocated new colour */
    lru = list->free;
    lruremove(lru);
    lru->col = col;
    lru->pixel = xcolor.pixel;
    lruinsert(&list->head, lru);
    return lru->pixel;
}
